@import url("bootstrap-astropy.css");

div.topbar a.brand {
  background: transparent url("logo_icon.png") no-repeat 8px 3px;
  background-image: url("logo_icon.png"), none;
  background-size: 32px 32px;
}

#logotext1 {
  color: #519EA8;
}

#logotext2 {
  color: #FF5000;
}
